from typing import List

from kirin import ir, types
from bloqade import qasm2
from kirin.rewrite import Walk, Fixpoint, CommonSubexpressionElimination
from kirin.dialects import py, func
from bloqade.qasm2.passes.glob import GlobalToUOP


def as_int(value: int):
    return py.constant.Constant(value=value)


def as_float(value: float):
    return py.constant.Constant(value=value)


def test_global_rewrite():

    @qasm2.extended
    def main():
        q1 = qasm2.qreg(1)
        q2 = qasm2.qreg(2)

        qasm2.glob.u(theta=1.3, phi=1.1, lam=1.2, registers=[q1, q2])

    GlobalToUOP(dialects=main.dialects)(main)

    main.print()

    # post-rewrite expected function
    expected: List[ir.Statement] = [
        (first_n_qubits := as_int(1)),
        (reg1 := qasm2.core.QRegNew(n_qubits=first_n_qubits.result)),
        (second_n_qubits := as_int(2)),
        (reg2 := qasm2.core.QRegNew(n_qubits=second_n_qubits.result)),
        (theta := as_float(1.3)),
        (phi := as_float(1.1)),
        (lam := as_float(1.2)),
        (idx0 := as_int(0)),
        (q0 := qasm2.core.QRegGet(reg1.result, idx=idx0.result)),
        (idx1 := as_int(1)),
        (q1 := qasm2.core.QRegGet(reg2.result, idx=idx1.result)),
        (idx2 := as_int(2)),
        (q2 := qasm2.core.QRegGet(reg2.result, idx=idx2.result)),
        (
            qasm2.uop.UGate(
                qarg=q2.result, theta=theta.result, phi=phi.result, lam=lam.result
            )
        ),
        (
            qasm2.uop.UGate(
                qarg=q1.result, theta=theta.result, phi=phi.result, lam=lam.result
            )
        ),
        (
            qasm2.uop.UGate(
                qarg=q0.result, theta=theta.result, phi=phi.result, lam=lam.result
            )
        ),
        (return_none := func.ConstantNone()),
        (func.Return(return_none)),
    ]
    block = ir.Block(expected)
    block.args.append_from(types.MethodType[[], types.NoneType], "main_self")
    expected_func_stmt = func.Function(
        sym_name="main",
        signature=func.Signature(inputs=(), output=types.NoneType),
        body=ir.Region(blocks=block),
    )

    expected_method = ir.Method(
        mod=None,
        py_func=None,
        sym_name="main",
        dialects=qasm2.main,
        code=expected_func_stmt,
        arg_names=[],
    )
    qasm2.main.run_pass(expected_method)
    Fixpoint(Walk(CommonSubexpressionElimination())).rewrite(expected_method.code)
    assert expected_method.code.is_equal(main.code)



import dataclasses
from typing import Any, Dict
from functools import cached_property

import numpy as np
import scipy.interpolate as interp

from .utils import array_to_json, json_to_array


@dataclasses.dataclass
class PPoly:
    c: np.ndarray
    x: np.ndarray
    extrapolate: bool = dataclasses.field(default=False)

    def __post_init__(self):
        super().__setattr__("x", np.asarray(self.x))
        super().__setattr__("c", np.asarray(self.c))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, PPoly):
            return False
        return (
            np.array_equal(self.x, other.x)
            and np.array_equal(self.c, other.c)
            and self.extrapolate == other.extrapolate
        )

    @cached_property
    def get_ppoly(self) -> interp.PPoly:
        return interp.PPoly(self.c, self.x, self.extrapolate)

    def __call__(self, x: float) -> float:
        return self.get_ppoly(x)

    def to_json(self) -> Dict[str, Any]:
        return {
            "c": array_to_json(self.c),
            "x": array_to_json(self.x),
            "extrapolate": self.extrapolate,
        }

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> "PPoly":
        return cls(
            c=json_to_array(json_dict["c"]),
            x=json_to_array(json_dict["x"]),
            extrapolate=json_dict["extrapolate"],
        )



import math
from typing import Union

from kirin.interp import Frame, Interpreter, MethodTable, impl

from . import stmts
from ._dialect import dialect


@dialect.register
class Qasm2UopInterpreter(MethodTable):
    name = "qasm2.uop"
    dialect = dialect

    @impl(stmts.ConstFloat)
    @impl(stmts.ConstInt)
    def new_const(
        self,
        interp: Interpreter,
        frame: Frame,
        stmt: Union[stmts.ConstFloat, stmts.ConstInt],
    ):
        return (stmt.value,)

    @impl(stmts.ConstPI)
    def new_const_pi(self, interp: Interpreter, frame: Frame, stmt: stmts.ConstPI):
        return (3.141592653589793,)

    @impl(stmts.Add)
    def add(self, interp: Interpreter, frame: Frame, stmt: stmts.Add):
        return (frame.get(stmt.lhs) + frame.get(stmt.rhs),)

    @impl(stmts.Sub)
    def sub(self, interp: Interpreter, frame: Frame, stmt: stmts.Sub):
        return (frame.get(stmt.lhs) - frame.get(stmt.rhs),)

    @impl(stmts.Mul)
    def mul(self, interp: Interpreter, frame: Frame, stmt: stmts.Mul):
        return (frame.get(stmt.lhs) * frame.get(stmt.rhs),)

    @impl(stmts.Div)
    def div(self, interp: Interpreter, frame: Frame, stmt: stmts.Div):
        return (frame.get(stmt.lhs) / frame.get(stmt.rhs),)

    @impl(stmts.Pow)
    def pow(self, interp: Interpreter, frame: Frame, stmt: stmts.Pow):
        return (frame.get(stmt.lhs) ** frame.get(stmt.rhs),)

    @impl(stmts.Neg)
    def neg(self, interp: Interpreter, frame: Frame, stmt: stmts.Neg):
        return (-frame.get(stmt.value),)

    @impl(stmts.Sqrt)
    def sqrt(self, interp: Interpreter, frame: Frame, stmt: stmts.Sqrt):
        return (math.sqrt(frame.get(stmt.value)),)

    @impl(stmts.Sin)
    def sin(self, interp: Interpreter, frame: Frame, stmt: stmts.Sin):
        return (math.sin(frame.get(stmt.value)),)

    @impl(stmts.Cos)
    def cos(self, interp: Interpreter, frame: Frame, stmt: stmts.Cos):
        return (math.cos(frame.get(stmt.value)),)

    @impl(stmts.Tan)
    def tan(self, interp: Interpreter, frame: Frame, stmt: stmts.Tan):
        return (math.tan(frame.get(stmt.value)),)

    @impl(stmts.Log)
    def log(self, interp: Interpreter, frame: Frame, stmt: stmts.Log):
        return (math.log(frame.get(stmt.value)),)

    @impl(stmts.Exp)
    def exp(self, interp: Interpreter, frame: Frame, stmt: stmts.Exp):
        return (math.exp(frame.get(stmt.value)),)



from typing import Union

from kirin.lowering import wraps

from .dialects import aux, gate, noise, collapse


# dialect:: gate
## 1q
@wraps(gate.X)
def x(targets: tuple[int, ...], dagger: bool = False) -> None: ...


@wraps(gate.Y)
def y(targets: tuple[int, ...], dagger: bool = False) -> None: ...


@wraps(gate.Z)
def z(targets: tuple[int, ...], dagger: bool = False) -> None: ...


@wraps(gate.Identity)
def identity(targets: tuple[int, ...], dagger: bool = False) -> None: ...


@wraps(gate.H)
def h(targets: tuple[int, ...], dagger: bool = False) -> None: ...


@wraps(gate.S)
def s(targets: tuple[int, ...], dagger: bool = False) -> None: ...


@wraps(gate.SqrtX)
def sqrt_x(targets: tuple[int, ...], dagger: bool = False) -> None: ...


@wraps(gate.SqrtY)
def sqrt_y(targets: tuple[int, ...], dagger: bool = False) -> None: ...


@wraps(gate.SqrtZ)
def sqrt_z(targets: tuple[int, ...], dagger: bool = False) -> None: ...


## clif 2q
@wraps(gate.Swap)
def swap(targets: tuple[int, ...], dagger: bool = False) -> None: ...


## ctrl 2q
@wraps(gate.CX)
def cx(
    controls: tuple[int, ...], targets: tuple[int, ...], dagger: bool = False
) -> None: ...


@wraps(gate.CY)
def cy(
    controls: tuple[int, ...], targets: tuple[int, ...], dagger: bool = False
) -> None: ...


@wraps(gate.CZ)
def cz(
    controls: tuple[int, ...], targets: tuple[int, ...], dagger: bool = False
) -> None: ...


## pp
@wraps(gate.SPP)
def spp(targets: tuple[aux.PauliString, ...], dagger=False) -> None: ...


# dialect:: aux
@wraps(aux.GetRecord)
def rec(id: int) -> aux.RecordResult: ...


@wraps(aux.Detector)
def detector(
    coord: tuple[Union[int, float], ...], targets: tuple[aux.RecordResult, ...]
) -> None: ...


@wraps(aux.ObservableInclude)
def observable_include(idx: int, targets: tuple[aux.RecordResult, ...]) -> None: ...


@wraps(aux.Tick)
def tick() -> None: ...


@wraps(aux.NewPauliString)
def pauli_string(
    string: tuple[str, ...], flipped: tuple[bool, ...], targets: tuple[int, ...]
) -> aux.PauliString: ...


# dialect:: collapse
@wraps(collapse.MZ)
def mz(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(collapse.MY)
def my(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(collapse.MX)
def mx(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(collapse.MZZ)
def mzz(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(collapse.MYY)
def myy(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(collapse.MXX)
def mxx(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(collapse.PPMeasurement)
def mpp(p: float, targets: tuple[aux.PauliString, ...]) -> None: ...


@wraps(collapse.RZ)
def rz(targets: tuple[int, ...]) -> None: ...


@wraps(collapse.RY)
def ry(targets: tuple[int, ...]) -> None: ...


@wraps(collapse.RX)
def rx(targets: tuple[int, ...]) -> None: ...


# dialect:: noise
@wraps(noise.Depolarize1)
def depolarize1(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(noise.Depolarize2)
def depolarize2(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(noise.PauliChannel1)
def pauli_channel1(
    px: float, py: float, pz: float, targets: tuple[int, ...]
) -> None: ...


@wraps(noise.PauliChannel2)
def pauli_channel2(
    pix: float,
    piy: float,
    piz: float,
    pxi: float,
    pxx: float,
    pxy: float,
    pxz: float,
    pyi: float,
    pyx: float,
    pyy: float,
    pyz: float,
    pzi: float,
    pzx: float,
    pzy: float,
    pzz: float,
    targets: tuple[int, ...],
) -> None: ...


@wraps(noise.XError)
def x_error(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(noise.YError)
def y_error(p: float, targets: tuple[int, ...]) -> None: ...


@wraps(noise.ZError)
def z_error(p: float, targets: tuple[int, ...]) -> None: ...



from io import StringIO
from typing import IO, TypeVar
from dataclasses import field, dataclass

from kirin import ir, interp
from kirin.emit import EmitStr, EmitStrFrame
from kirin.dialects import func

IO_t = TypeVar("IO_t", bound=IO)


def _default_dialect_group() -> ir.DialectGroup:
    from ..groups import main

    return main


@dataclass
class EmitStimMain(EmitStr):
    keys = ["emit.stim"]
    dialects: ir.DialectGroup = field(default_factory=_default_dialect_group)
    file: StringIO = field(default_factory=StringIO)

    def initialize(self):
        super().initialize()
        self.file.truncate(0)
        self.file.seek(0)
        return self

    def eval_stmt_fallback(
        self, frame: EmitStrFrame, stmt: ir.Statement
    ) -> tuple[str, ...]:
        return (stmt.name,)

    def emit_block(self, frame: EmitStrFrame, block: ir.Block) -> str | None:
        for stmt in block.stmts:
            result = self.eval_stmt(frame, stmt)
            if isinstance(result, tuple):
                frame.set_values(stmt.results, result)
        return None

    def get_output(self) -> str:
        self.file.seek(0)
        return self.file.read()


@func.dialect.register(key="emit.stim")
class FuncEmit(interp.MethodTable):

    @interp.impl(func.Function)
    def emit_func(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: func.Function):
        _ = emit.run_ssacfg_region(frame, stmt.body)
        # emit.output = "\n".join(frame.body)
        return ()



from kirin import ir, types
from kirin.decl import info, statement

from .._dialect import dialect
from ...aux.types import PauliStringType


# Generalized Pauli-product gates
# ---------------------------------------
@statement(dialect=dialect)
class SPP(ir.Statement):
    name = "SPP"
    traits = frozenset({ir.FromPythonCall()})
    dagger: bool = info.attribute(types.Bool, default=False)
    targets: tuple[ir.SSAValue, ...] = info.argument(PauliStringType)



from kirin import types
from bloqade.types import Qubit as Qubit, QubitType as QubitType


class Bit:
    """Runtime representation of a bit.

    Note:
        This is the base class of more specific bit types, such as
        a reference to a piece of classical register in some quantum register
        dialects.
    """

    pass


class QReg:
    """Runtime representation of a quantum register."""

    def __getitem__(self, index) -> Qubit:
        raise NotImplementedError("cannot call __getitem__ outside of a kernel")


class CReg:
    """Runtime representation of a classical register."""

    def __getitem__(self, index) -> Bit:
        raise NotImplementedError("cannot call __getitem__ outside of a kernel")


BitType = types.PyClass(Bit)
"""Kirin type for a classical bit."""

QRegType = types.PyClass(QReg)
"""Kirin type for a quantum register."""

CRegType = types.PyClass(CReg)
"""Kirin type for a classical register."""



from .global_to_uop import GlobalToUOpRule as GlobalToUOpRule
from .parallel_to_uop import ParallelToUOpRule as ParallelToUOpRule
from .uop_to_parallel import (
    MergePolicyABC as MergePolicyABC,
    UOpToParallelRule as UOpToParallelRule,
    SimpleGreedyMergePolicy as SimpleGreedyMergePolicy,
    SimpleOptimalMergePolicy as SimpleOptimalMergePolicy,
)



from dataclasses import dataclass

from kirin import ir, lowering
from kirin.dialects import cf, func, ilist
from kirin.lowering import LoweringState
from kirin.exceptions import DialectLoweringError
from bloqade.qasm2.types import CRegType, QRegType
from bloqade.qasm2.dialects import uop, core, expr, glob, noise, parallel

from . import ast
from .visitor import Visitor


@dataclass
class LoweringQASM(Visitor[lowering.Result]):
    state: LoweringState
    extension: str | None = None

    def visit_MainProgram(self, node: ast.MainProgram) -> lowering.Result:
        allowed = {dialect.name for dialect in self.state.dialects}
        if isinstance(node.header, ast.OPENQASM) and node.header.version.major == 2:
            dialects = ["qasm2.core", "qasm2.uop", "qasm2.expr"]
        elif isinstance(node.header, ast.Kirin):
            dialects = node.header.dialects

        for dialect in dialects:
            if dialect not in allowed:
                raise DialectLoweringError(
                    f"Dialect {dialect} not found, allowed: {', '.join(allowed)}"
                )

        for stmt in node.statements:
            self.visit(stmt)
        return lowering.Result()

    def visit_QReg(self, node: ast.QReg) -> lowering.Result:
        reg = core.QRegNew(
            self.state.append_stmt(expr.ConstInt(value=node.size)).result
        )
        self.state.append_stmt(reg)
        self.state.current_frame.defs[node.name] = reg.result
        return lowering.Result()

    def visit_CReg(self, node: ast.CReg) -> lowering.Result:
        reg = core.CRegNew(
            self.state.append_stmt(expr.ConstInt(value=node.size)).result
        )
        self.state.append_stmt(reg)
        self.state.current_frame.defs[node.name] = reg.result
        return lowering.Result()

    def visit_Barrier(self, node: ast.Barrier) -> lowering.Result:
        self.state.append_stmt(
            uop.Barrier(
                qargs=tuple(self.visit(qarg).expect_one() for qarg in node.qargs)
            )
        )
        return lowering.Result()

    def visit_CXGate(self, node: ast.CXGate) -> lowering.Result:
        self.state.append_stmt(
            uop.CX(
                ctrl=self.visit(node.ctrl).expect_one(),
                qarg=self.visit(node.qarg).expect_one(),
            )
        )
        return lowering.Result()

    def visit_Measure(self, node: ast.Measure) -> lowering.Result:
        self.state.append_stmt(
            core.Measure(
                qarg=self.visit(node.qarg).expect_one(),
                carg=self.visit(node.carg).expect_one(),
            )
        )
        return lowering.Result()

    def visit_UGate(self, node: ast.UGate) -> lowering.Result:
        self.state.append_stmt(
            uop.UGate(
                theta=self.visit(node.theta).expect_one(),
                phi=self.visit(node.phi).expect_one(),
                lam=self.visit(node.lam).expect_one(),
                qarg=self.visit(node.qarg).expect_one(),
            )
        )
        return lowering.Result()

    def visit_Reset(self, node: ast.Reset) -> lowering.Result:
        self.state.append_stmt(core.Reset(qarg=self.visit(node.qarg).expect_one()))
        return lowering.Result()

    def visit_IfStmt(self, node: ast.IfStmt) -> lowering.Result:
        cond_stmt = core.CRegEq(
            lhs=self.visit(node.cond.lhs).expect_one(),
            rhs=self.visit(node.cond.rhs).expect_one(),
        )
        cond = self.state.append_stmt(cond_stmt).result
        frame = self.state.current_frame
        before_block = frame.curr_block
        if frame.exit_block is None:
            raise DialectLoweringError("code block is not exiting")
        else:
            before_block_next = frame.exit_block
        if_block = self.state.current_frame.append_block()
        for stmt in node.body:
            self.visit(stmt)

        if_block.stmts.append(
            cf.Branch(
                arguments=(),
                successor=before_block_next,
            )
        )
        before_block.stmts.append(
            cf.ConditionalBranch(
                cond=cond,
                then_arguments=(),
                then_successor=if_block,
                else_arguments=(),
                else_successor=before_block_next,
            )
        )
        frame.curr_block = before_block_next
        return lowering.Result()

    def visit_BinOp(self, node: ast.BinOp) -> lowering.Result:
        if node.op == "+":
            stmt_type = expr.Add
        elif node.op == "-":
            stmt_type = expr.Sub
        elif node.op == "*":
            stmt_type = expr.Mul
        else:
            stmt_type = expr.Div

        stmt = self.state.append_stmt(
            stmt_type(
                lhs=self.visit(node.lhs).expect_one(),
                rhs=self.visit(node.rhs).expect_one(),
            )
        )
        return lowering.Result(stmt.result)

    def visit_UnaryOp(self, node: ast.UnaryOp) -> lowering.Result:
        if node.op == "-":
            stmt = expr.Neg(value=self.visit(node.operand).expect_one())
            return lowering.Result(stmt.result)
        else:
            return lowering.Result(self.visit(node.operand).expect_one())

    def visit_Bit(self, node: ast.Bit) -> lowering.Result:
        if node.name.id not in self.state.current_frame.defs:
            raise ValueError(f"Bit {node.name} not found")

        addr = self.state.append_stmt(expr.ConstInt(value=node.addr))
        reg = self.state.current_frame.get_local(node.name.id)
        if reg is None:
            raise DialectLoweringError(f"{node.name.id} is not defined")

        if reg.type.is_subseteq(QRegType):
            stmt = core.QRegGet(reg, addr.result)
        elif reg.type.is_subseteq(CRegType):
            stmt = core.CRegGet(reg, addr.result)
        return lowering.Result(self.state.append_stmt(stmt).result)

    def visit_Call(self, node: ast.Call) -> lowering.Result:
        if node.name == "cos":
            stmt = expr.Cos(self.visit(node.args[0]).expect_one())
        elif node.name == "sin":
            stmt = expr.Sin(self.visit(node.args[0]).expect_one())
        elif node.name == "tan":
            stmt = expr.Tan(self.visit(node.args[0]).expect_one())
        elif node.name == "exp":
            stmt = expr.Exp(self.visit(node.args[0]).expect_one())
        elif node.name == "log":
            stmt = expr.Log(self.visit(node.args[0]).expect_one())
        elif node.name == "sqrt":
            stmt = expr.Sqrt(self.visit(node.args[0]).expect_one())
        else:
            raise ValueError(f"Unknown function {node.name}")
        self.state.append_stmt(stmt)
        return lowering.Result(stmt.result)

    def visit_Instruction(self, node: ast.Instruction) -> lowering.Result:
        params = [self.visit(param).expect_one() for param in node.params]
        qargs = [self.visit(qarg).expect_one() for qarg in node.qargs]
        visit_inst = getattr(self, "visit_instruction_" + node.name.id, None)
        if visit_inst is not None:
            self.state.append_stmt(visit_inst(node, params, qargs))
        else:
            value = self.state.get_global(node.name.id).expect(ir.Method)
            # NOTE: QASM expects the return type to be known at call site
            if value.return_type is None:
                raise ValueError(f"Unknown return type for {node.name.id}")
            self.state.append_stmt(
                func.Invoke(
                    callee=value,
                    inputs=tuple(params + qargs),
                    kwargs=tuple(),
                )
            )
        return lowering.Result()

    def visit_instruction_id(self, node: ast.Instruction, params, qargs):
        return uop.Id(qarg=qargs[0])

    def visit_instruction_x(self, node: ast.Instruction, params, qargs):
        return uop.X(qarg=qargs[0])

    def visit_instruction_y(self, node: ast.Instruction, params, qargs):
        return uop.Y(qarg=qargs[0])

    def visit_instruction_z(self, node: ast.Instruction, params, qargs):
        return uop.Z(qarg=qargs[0])

    def visit_instruction_h(self, node: ast.Instruction, params, qargs):
        return uop.H(qarg=qargs[0])

    def visit_instruction_s(self, node: ast.Instruction, params, qargs):
        return uop.S(qarg=qargs[0])

    def visit_instruction_sdg(self, node: ast.Instruction, params, qargs):
        return uop.Sdag(qarg=qargs[0])

    def visit_instruction_sx(self, node: ast.Instruction, params, qargs):
        return uop.SX(qarg=qargs[0])

    def visit_instruction_sxdg(self, node: ast.Instruction, params, qargs):
        return uop.SXdag(qarg=qargs[0])

    def visit_instruction_t(self, node: ast.Instruction, params, qargs):
        return uop.T(qarg=qargs[0])

    def visit_instruction_tdg(self, node: ast.Instruction, params, qargs):
        return uop.Tdag(qarg=qargs[0])

    def visit_instruction_rx(self, node: ast.Instruction, params, qargs):
        return uop.RX(theta=params[0], qarg=qargs[0])

    def visit_instruction_ry(self, node: ast.Instruction, params, qargs):
        return uop.RY(theta=params[0], qarg=qargs[0])

    def visit_instruction_rz(self, node: ast.Instruction, params, qargs):
        return uop.RZ(theta=params[0], qarg=qargs[0])

    def visit_instruction_p(self, node: ast.Instruction, params, qargs):
        return uop.U1(lam=params[0], qarg=qargs[0])

    def visit_instruction_u(self, node: ast.Instruction, params, qargs):
        return uop.UGate(theta=params[0], phi=params[1], lam=params[2], qarg=qargs[0])

    def visit_instruction_u1(self, node: ast.Instruction, params, qargs):
        return uop.U1(lam=params[0], qarg=qargs[0])

    def visit_instruction_u2(self, node: ast.Instruction, params, qargs):
        return uop.U2(phi=params[0], lam=params[1], qarg=qargs[0])

    def visit_instruction_u3(self, node: ast.Instruction, params, qargs):
        return uop.UGate(theta=params[0], phi=params[1], lam=params[2], qarg=qargs[0])

    def visit_instruction_CX(self, node: ast.Instruction, params, qargs):
        return uop.CX(ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_cx(self, node: ast.Instruction, params, qargs):
        return uop.CX(ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_cy(self, node: ast.Instruction, params, qargs):
        return uop.CY(ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_cz(self, node: ast.Instruction, params, qargs):
        return uop.CZ(ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_ch(self, node: ast.Instruction, params, qargs):
        return uop.CH(ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_crx(self, node: ast.Instruction, params, qargs):
        return uop.CRX(lam=params[0], ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_cry(self, node: ast.Instruction, params, qargs):
        return uop.CRY(lam=params[0], ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_crz(self, node: ast.Instruction, params, qargs):
        return uop.CRZ(lam=params[0], ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_ccx(self, node: ast.Instruction, params, qargs):
        return uop.CCX(ctrl1=qargs[0], ctrl2=qargs[1], qarg=qargs[2])

    def visit_instruction_csx(self, node: ast.Instruction, params, qargs):
        return uop.CSX(ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_cswap(self, node: ast.Instruction, params, qargs):
        return uop.CSwap(ctrl=qargs[0], qarg1=qargs[1], qarg2=qargs[2])

    def visit_instruction_cp(self, node: ast.Instruction, params, qargs):
        return uop.CU1(lam=params[0], ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_cu1(self, node: ast.Instruction, params, qargs):
        return uop.CU1(lam=params[0], ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_cu3(self, node: ast.Instruction, params, qargs):
        return uop.CU3(
            theta=params[0], phi=params[1], lam=params[2], ctrl=qargs[0], qarg=qargs[1]
        )

    def visit_instruction_cu(self, node: ast.Instruction, params, qargs):
        return uop.CU3(
            theta=params[0], phi=params[1], lam=params[2], ctrl=qargs[0], qarg=qargs[1]
        )

    def visit_instruction_rxx(self, node: ast.Instruction, params, qargs):
        return uop.RXX(theta=params[0], ctrl=qargs[0], qarg=qargs[1])

    def visit_instruction_rzz(self, node: ast.Instruction, params, qargs):
        return uop.RZZ(theta=params[0], ctrl=qargs[0], qarg=qargs[1])

    def visit_Number(self, node: ast.Number) -> lowering.Result:
        if isinstance(node.value, int):
            stmt = expr.ConstInt(value=node.value)
        else:
            stmt = expr.ConstFloat(value=node.value)
        return lowering.Result(self.state.append_stmt(stmt).result)

    def visit_Pi(self, node: ast.Pi) -> lowering.Result:
        return lowering.Result(self.state.append_stmt(expr.ConstPI()).result)

    def visit_Include(self, node: ast.Include) -> lowering.Result:
        if node.filename not in ["qelib1.inc"]:
            raise DialectLoweringError(f"Include {node.filename} not found")

        return lowering.Result()

    def visit_Gate(self, node: ast.Gate) -> lowering.Result:
        raise NotImplementedError("Gate lowering not supported")

    def visit_Name(self, node: ast.Name) -> lowering.Result:
        if (value := self.state.current_frame.get_local(node.id)) is not None:
            return lowering.Result(value)
        raise ValueError(f"name {node.id} not found")

    def visit_ParaCZGate(self, node: ast.ParaCZGate) -> lowering.Result:
        ctrls: list[ir.SSAValue] = []
        qargs: list[ir.SSAValue] = []
        for pair in node.qargs:
            if len(pair) != 2:
                raise ValueError("CZ gate requires exactly two qargs")
            ctrl, qarg = pair
            ctrls.append(self.visit(ctrl).expect_one())
            qargs.append(self.visit(qarg).expect_one())

        ctrls_stmt = ilist.New(values=ctrls)
        qargs_stmt = ilist.New(values=qargs)
        self.state.append_stmt(ctrls_stmt)
        self.state.append_stmt(qargs_stmt)
        self.state.append_stmt(
            parallel.CZ(ctrls=ctrls_stmt.result, qargs=qargs_stmt.result)
        )
        return lowering.Result()

    def visit_ParaRZGate(self, node: ast.ParaRZGate) -> lowering.Result:
        qargs: list[ir.SSAValue] = []
        for pair in node.qargs:
            if len(pair) != 1:
                raise ValueError("Rz gate requires exactly one qarg")
            qargs.append(self.visit(pair[0]).expect_one())

        qargs_stmt = ilist.New(values=qargs)
        self.state.append_stmt(qargs_stmt)
        self.state.append_stmt(
            parallel.RZ(
                theta=self.visit(node.theta).expect_one(),
                qargs=qargs_stmt.result,
            )
        )
        return lowering.Result()

    def visit_ParaU3Gate(self, node: ast.ParaU3Gate) -> lowering.Result:
        qargs: list[ir.SSAValue] = []
        for pair in node.qargs:
            if len(pair) != 1:
                raise ValueError("U3 gate requires exactly one qarg")
            qargs.append(self.visit(pair[0]).expect_one())

        qargs_stmt = ilist.New(values=qargs)
        self.state.append_stmt(qargs_stmt)
        self.state.append_stmt(
            parallel.UGate(
                theta=self.visit(node.theta).expect_one(),
                phi=self.visit(node.phi).expect_one(),
                lam=self.visit(node.lam).expect_one(),
                qargs=qargs_stmt.result,
            )
        )
        return lowering.Result()

    def visit_GlobUGate(self, node: ast.GlobUGate) -> lowering.Result:

        registers: list[ir.SSAValue] = []

        for register in node.registers:  # These will all be ast.Names
            registers.append(self.visit(register).expect_one())

        registers_stmt = ilist.New(values=registers)
        self.state.append_stmt(registers_stmt)
        self.state.append_stmt(
            # all the stuff going into the args should be SSA values
            glob.UGate(
                registers=registers_stmt.result,  # expect_one = a singular SSA value
                theta=self.visit(node.theta).expect_one(),
                phi=self.visit(node.phi).expect_one(),
                lam=self.visit(node.lam).expect_one(),
            )
        )
        return lowering.Result()

    def visit_NoisePAULI1(self, node: ast.NoisePAULI1) -> lowering.Result:
        self.state.append_stmt(
            noise.Pauli1(
                px=self.visit(node.px).expect_one(),
                py=self.visit(node.py).expect_one(),
                pz=self.visit(node.pz).expect_one(),
                qarg=self.visit(node.qarg).expect_one(),
            )
        )
        return lowering.Result()



from dataclasses import dataclass

from kirin import ir, interp
from kirin.dialects import cf, scf, func
from kirin.ir.dialect import Dialect as Dialect
from bloqade.qasm2.parse import ast

from .base import EmitQASM2Base, EmitQASM2Frame


@dataclass
class EmitQASM2Main(EmitQASM2Base[ast.Statement, ast.MainProgram]):
    keys = ["emit.qasm2.main", "emit.qasm2.gate"]
    dialects: ir.DialectGroup


@func.dialect.register(key="emit.qasm2.main")
class Func(interp.MethodTable):

    @interp.impl(func.Function)
    def emit_func(
        self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: func.Function
    ):
        from bloqade.qasm2.dialects import glob, noise, parallel

        emit.run_ssacfg_region(frame, stmt.body)
        if emit.dialects.data.intersection(
            (parallel.dialect, glob.dialect, noise.dialect)
        ):
            header = ast.Kirin([dialect.name for dialect in emit.dialects])
        else:
            header = ast.OPENQASM(ast.Version(2, 0))

        emit.output = ast.MainProgram(header=header, statements=frame.body)
        return ()


@cf.dialect.register(key="emit.qasm2.main")
class Cf(interp.MethodTable):

    @interp.impl(cf.Branch)
    def emit_branch(self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: cf.Branch):
        frame.worklist.append(
            interp.Successor(stmt.successor, frame.get_values(stmt.arguments))
        )
        return ()

    @interp.impl(cf.ConditionalBranch)
    def emit_conditional_branch(
        self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: cf.ConditionalBranch
    ):
        cond = emit.assert_node(ast.Cmp, frame.get(stmt.cond))
        body_frame = emit.new_frame(stmt)
        body_frame.entries.update(frame.entries)
        body_frame.set_values(
            stmt.then_successor.args, frame.get_values(stmt.then_arguments)
        )
        emit.emit_block(body_frame, stmt.then_successor)
        frame.body.append(
            ast.IfStmt(
                cond,
                body=body_frame.body,  # type: ignore
            )
        )
        frame.worklist.append(
            interp.Successor(stmt.else_successor, frame.get_values(stmt.else_arguments))
        )
        return ()


@scf.dialect.register(key="emit.qasm2.main")
class Scf(interp.MethodTable):

    @interp.impl(scf.Yield)
    def emit_yield(self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: scf.Yield):
        return frame.get_values(stmt.values)

    @interp.impl(scf.IfElse)
    def emit_if_else(
        self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: scf.IfElse
    ):
        else_stmts = stmt.else_body.blocks[0].stmts
        if not (
            len(else_stmts) == 0
            or len(else_stmts) == 1
            and isinstance(else_stmts.at(0), scf.Yield)
        ):
            raise interp.InterpreterError(
                "cannot lower if-else with non-empty else block"
            )

        cond = emit.assert_node(ast.Cmp, frame.get(stmt.cond))
        then_frame = emit.new_frame(stmt)
        then_frame.entries.update(frame.entries)
        emit.emit_block(then_frame, stmt.then_body.blocks[0])
        frame.body.append(
            ast.IfStmt(
                cond,
                body=then_frame.body,  # type: ignore
            )
        )
        term = stmt.then_body.blocks[0].last_stmt
        if isinstance(term, scf.Yield):
            return then_frame.get_values(term.values)
        return ()



from kirin import interp
from bloqade.analysis.address import (
    Address,
    NotQubit,
    AddressReg,
    AddressQubit,
    AddressAnalysis,
)

from .stmts import QRegGet, QRegNew
from ._dialect import dialect


@dialect.register(key="qubit.address")
class AddressMethodTable(interp.MethodTable):

    @interp.impl(QRegNew)
    def new(
        self,
        interp: AddressAnalysis,
        frame: interp.Frame[Address],
        stmt: QRegNew,
    ):
        n_qubits = interp.get_const_value(int, stmt.n_qubits)
        addr = AddressReg(range(interp.next_address, interp.next_address + n_qubits))
        interp.next_address += n_qubits
        return (addr,)

    @interp.impl(QRegGet)
    def get(self, interp: AddressAnalysis, frame: interp.Frame[Address], stmt: QRegGet):
        addr = frame.get(stmt.reg)
        pos = interp.get_const_value(int, stmt.idx)
        if isinstance(addr, AddressReg):
            return (AddressQubit(addr.data[pos]),)
        else:  # this is not reachable
            return (NotQubit(),)



from kirin.lowering import wraps

from .types import Bit, CReg, QReg, Qubit
from .dialects import uop, core, expr, inline as inline_


@wraps(inline_.InlineQASM)
def inline(text: str) -> None:
    """
    Inline QASM code into the current program.

    Args:
        text: The QASM code to inline.
    """
    ...


@wraps(core.QRegNew)
def qreg(n_qubits: int) -> QReg:
    """
    Create a new quantum register with `n_qubits` qubits.

    Args:
        n_qubits: The number of qubits in the register.

    Returns:
        The newly created quantum register.

    """
    ...


@wraps(core.CRegNew)
def creg(n_bits: int) -> CReg:
    """
    Create a new classical register with `n_bits` bits.

    Args:
        n_bits: The number of bits in the register.

    Returns:
        The newly created classical register.

    """
    ...


@wraps(core.Reset)
def reset(qarg: Qubit) -> None:
    """
    Reset the qubit `qarg` to the |0⟩ state.

    Args:
        qarg: The qubit to reset.

    """

    ...


@wraps(core.Measure)
def measure(qarg: Qubit, cbit: Bit) -> None:
    """
    Measure the qubit `qarg` and store the result in the classical bit `cbit`.

    Args:
        qarg: The qubit to measure.
        cbit: The classical bit to store the result in.
    """
    ...


@wraps(uop.CX)
def cx(ctrl: Qubit, qarg: Qubit) -> None:
    """
    Controlled-X (CNOT) gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit.
    """
    ...


@wraps(uop.UGate)
def u(qarg: Qubit, theta: float, phi: float, lam: float) -> None:
    """
    U gate.

    Note:
        See https://arxiv.org/pdf/1707.03429 for definition of angles.

    Args:
        qarg: The qubit to apply the gate to.
        theta: The angle of rotation
        phi: The angle of rotation
        lam: The angle of rotation

    """
    ...


@wraps(uop.UGate)
def u3(qarg: Qubit, theta: float, phi: float, lam: float) -> None:
    """
    U3 gate, same as u

    Note:
        See https://arxiv.org/pdf/1707.03429 for definition of angles.

    Args:
        qarg: The qubit to apply the gate to.
        theta: The angle of rotation
        phi: The angle of rotation
        lam: The angle of rotation

    """
    ...


@wraps(uop.Barrier)
def barrier(qargs: tuple[Qubit, ...]) -> None:
    """
    Barrier instruction.

    Args:
        qargs: The qubits to apply the barrier to.
    """

    ...


@wraps(uop.Id)
def id(qarg: Qubit) -> None:
    """
    Identity gate.

    Args:
        qarg: The qubit to apply the gate to.

    """
    ...


@wraps(uop.H)
def h(qarg: Qubit) -> None:
    """
    Hadamard gate.

    Args:
        qarg: The qubit to apply the gate to.

    """
    ...


@wraps(uop.X)
def x(qarg: Qubit) -> None:
    """
    Pauli-X gate.

    Args:
        qarg: The qubit to apply the gate to.
    """

    ...


@wraps(uop.Y)
def y(qarg: Qubit) -> None:
    """
    Pauli-Y gate.

    Args:
        qarg: The qubit to apply the gate to.

    """
    ...


@wraps(uop.Z)
def z(qarg: Qubit) -> None:
    """
    Pauli-Z gate.

    Args:
        qarg: The qubit to apply the gate to.

    """
    ...


@wraps(uop.U1)
def p(qarg: Qubit, lam: float) -> None:
    """
    Phase gate.

    This is equivalent to u(0,0,lam), and u1(lam)

    Args:
        qarg: The qubit to apply the gate to.
        lam: The angle of phase.

    """
    ...


@wraps(uop.S)
def s(qarg: Qubit) -> None:
    """
    S gate.

    Args:
        qarg: The qubit to apply the gate to.
    """

    ...


@wraps(uop.Sdag)
def sdg(qarg: Qubit) -> None:
    """
    Hermitian conjugate of the S gate.

    Args:
        qarg: The qubit to apply the gate to.

    """

    ...


@wraps(uop.SX)
def sx(qarg: Qubit) -> None:
    """
    Sqrt(X) gate.

    Args:
        qarg: The qubit to apply the gate to.
    """

    ...


@wraps(uop.SXdag)
def sxdg(qarg: Qubit) -> None:
    """
    Hermitian conjugate of Sqrt(X) gate.

    Args:
        qarg: The qubit to apply the gate to.
    """

    ...


@wraps(uop.T)
def t(qarg: Qubit) -> None:
    """
    T gate.

    Args:
        qarg: The qubit to apply the gate to.
    """

    ...


@wraps(uop.Tdag)
def tdg(qarg: Qubit) -> None:
    """
    Hermitian conjugate of the T gate.

    Args:
        qarg: The qubit to apply the gate to.

    """

    ...


@wraps(uop.RX)
def rx(qarg: Qubit, theta: float) -> None:
    """
    Single qubit rotation about the X axis on block sphere

    Args:
        qarg: The qubit to apply the gate to.
        theta: The angle of rotation.
    """
    ...


@wraps(uop.RY)
def ry(qarg: Qubit, theta: float) -> None:
    """
    Single qubit rotation about the Y axis on block sphere

    Args:
        qarg: The qubit to apply the gate to.
        theta: The angle of rotation.

    """

    ...


@wraps(uop.RZ)
def rz(qarg: Qubit, theta: float) -> None:
    """
    Single qubit rotation about the Z axis on block sphere

    Args:
        qarg: The qubit to apply the gate to.
        theta: The angle of rotation.
    """
    ...


@wraps(uop.U1)
def u1(qarg: Qubit, lam: float) -> None:
    """
    1 Parameter single qubit unitary gate.

    This is equivalent to u(0,0,lambda).

    Args:
        qarg: The qubit to apply the gate to.
        lam: The angle of rotation.
    """
    ...


@wraps(uop.U2)
def u2(qarg: Qubit, phi: float, lam: float) -> None:
    """
    2 Parameter single qubit unitary gate.

    This is equivalent to u(pi/2,phi,lambda)

    Args:
        qarg: The qubit to apply the gate to.
        phi: The angle of rotation.
        lam: The angle of rotation.
    """
    ...


@wraps(uop.CZ)
def cz(ctrl: Qubit, qarg: Qubit) -> None:
    """
    Controlled-Z gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit
    """
    ...


@wraps(uop.CSX)
def csx(ctrl: Qubit, qarg: Qubit) -> None:
    """
    Controlled-Sqrt(X) gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit
    """
    ...


@wraps(uop.CY)
def cy(ctrl: Qubit, qarg: Qubit) -> None:
    """
    Controlled-Y gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit
    """

    ...


@wraps(uop.CH)
def ch(ctrl: Qubit, qarg: Qubit) -> None:
    """
    Controlled-Hadamard gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit

    """

    ...


@wraps(uop.CCX)
def ccx(ctrl1: Qubit, ctrl2: Qubit, qarg: Qubit) -> None:
    """
    Toffoli gate.

    Args:
        ctrl1: The first control qubit.
        ctrl2: The second control qubit.
        qarg: The target qubit.
    """
    ...


@wraps(uop.CSwap)
def cswap(ctrl: Qubit, qarg1: Qubit, qarg2: Qubit) -> None:
    """
    Controlled Swap gate (Fredkin gate).

    Args:
        ctrl: The control qubit.
        qarg1: The first target qubit.
        qarg2: The second target qubit.
    """
    ...


@wraps(uop.CRX)
def crx(ctrl: Qubit, qarg: Qubit, lam: float) -> None:
    """
    Controlled Rx rotation gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit.
        lam: The angle of rotation.

    """

    ...


@wraps(uop.CRY)
def cry(ctrl: Qubit, qarg: Qubit, lam: float) -> None:
    """
    Controlled Ry rotation gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit.
        lam: The angle of rotation.

    """

    ...


@wraps(uop.CRZ)
def crz(ctrl: Qubit, qarg: Qubit, lam: float) -> None:
    """
    Controlled Rz rotation gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit.
        lam: The angle of rotation.

    """
    ...


@wraps(uop.CU1)
def cu1(ctrl: Qubit, qarg: Qubit, lam: float) -> None:
    """
    Controlled phase rotation gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit.
        lam: The angle of rotation.
    """

    ...


@wraps(uop.CU1)
def cp(ctrl: Qubit, qarg: Qubit, lam: float) -> None:
    """
    Controlled phase rotation gate. Same as cu1

    Args:
        ctrl: The control qubit.
        qarg: The target qubit.
        lam: The angle of rotation.
    """

    ...


@wraps(uop.CU3)
def cu3(ctrl: Qubit, qarg: Qubit, theta: float, phi: float, lam: float) -> None:
    """
    Controlled 3-parameter unitary gate.

    Args:
        ctrl: The control qubit.
        qarg: The target qubit.
        theta: The angle of rotation.
        phi: The angle of rotation.
        lam: The angle of rotation.

    """
    ...


@wraps(uop.CU)
def cu(
    ctrl: Qubit, qarg: Qubit, theta: float, phi: float, lam: float, gamma: float
) -> None:
    """
    Controlled 4-parameter unitary gate.

    This is equal to:

    gate cu(theta,phi,lambda,gamma) c, t{
        p(gamma) c;
        p((lambda+phi)/2) c;
        p((lambda-phi)/2) t;
        cx c,t;
        u(-theta/2,0,-(phi+lambda)/2) t;
        cx c,t;
        u(theta/2,phi,0) t;
    }

    Args:
        ctrl: The control qubit.
        qarg: The target qubit.
        theta: The angle of rotation.
        phi: The angle of rotation.
        lam: The angle of rotation.
        gamma: The angle of rotation.
    """
    ...


@wraps(uop.RXX)
def rxx(ctrl: Qubit, qarg: Qubit, theta: float) -> None:
    """
    XX rotation gate.

    Args:
        ctrl: The first qubit.
        qarg: The second qubit.
        theta: The angle of rotation.

    """
    ...


@wraps(uop.RZZ)
def rzz(ctrl: Qubit, qarg: Qubit, theta: float) -> None:
    """
    ZZ rotation gate.

    Args:
        ctrl: The first qubit.
        qarg: The second qubit.
        theta: The angle of rotation.

    """
    ...


@wraps(expr.Sin)
def sin(value: float) -> float:
    """
    Sine math function.

    Args:
        value: The value to take the sine of.

    Returns:
        The sine of `value`.

    """
    ...


@wraps(expr.Cos)
def cos(value: float) -> float:
    """
    Cosine math function.

    Args:
        value: The value to take the cosine of.

    Returns:
        The cosine of `value`.

    """

    ...


@wraps(expr.Tan)
def tan(value: float) -> float:
    """
    Tangent math function.

    Args:
        value: The value to take the tangent of.

    Returns:
        The tangent of `value`.

    """

    ...


@wraps(expr.Exp)
def exp(value: float) -> float:
    """
    Exponential math function.

    Args:
        value: The value to exponentiate.

    Returns:
        The exponential of `value`.

    """

    ...


@wraps(expr.Log)
def ln(value: float) -> float:
    """
    logarithm math function.

    Args:
        value: The value to take the natural logarithm of.

    Returns:
        The natural logarithm of `value`.

    """

    ...


@wraps(expr.Sqrt)
def sqrt(value: float) -> float:
    """
    Square root math function.

    Args:
        value: The value to take the square root of.

    Returns:
        The square root of `value`.
    """
    ...



from kirin import passes
from bloqade import qasm2
from kirin.dialects import py, ilist
from bloqade.analysis.address import (
    NotQubit,
    AddressReg,
    AnyAddress,
    AddressQubit,
    AddressTuple,
    AddressAnalysis,
)

address = AddressAnalysis(qasm2.main.add(py.tuple).add(ilist))
fold = passes.Fold(qasm2.main.add(py.tuple).add(ilist))


def test_fixed_count():
    @qasm2.main
    def fixed_count():
        ra = qasm2.qreg(3)
        rb = qasm2.qreg(4)
        q0 = ra[0]
        q3 = rb[1]
        qasm2.x(q0)
        qasm2.x(q3)
        return q3

    fold(fixed_count)
    results, ret = address.run_analysis(fixed_count)
    # fixed_count.print(analysis=address.results)
    assert isinstance(ret, AddressQubit)
    assert ret.data == range(3, 7)[1]
    assert address.qubit_count == 7


def test_multiple_return_only_reg():

    @qasm2.main.add(py.tuple)
    def tuple_count():
        ra = qasm2.qreg(3)
        rb = qasm2.qreg(4)
        return ra, rb

    # tuple_count.dce()
    fold(tuple_count)
    frame, ret = address.run_analysis(tuple_count)
    # tuple_count.code.print(analysis=frame.entries)
    assert isinstance(ret, AddressTuple)
    assert isinstance(ret.data[0], AddressReg) and ret.data[0].data == range(0, 3)
    assert isinstance(ret.data[1], AddressReg) and ret.data[1].data == range(3, 7)


def test_dynamic_address():
    @qasm2.main
    def dynamic_address():
        ra = qasm2.qreg(3)
        rb = qasm2.qreg(4)
        ca = qasm2.creg(2)
        qasm2.measure(ra[0], ca[0])
        qasm2.measure(rb[1], ca[1])
        if ca[0] == ca[1]:
            return ra
        else:
            return rb

    # dynamic_address.code.print()
    dynamic_address.print()
    fold(dynamic_address)
    frame, result = address.run_analysis(dynamic_address)
    assert isinstance(result, AnyAddress)


# NOTE: this is invalid for QASM2
# def test_cond_count2():
#     @qasm2.main
#     def cond_count2():
#         ra = qasm2.qreg(3)
#         rb = qasm2.qreg(4)
#         if 4 > 33:
#             return 3.0
#         else:
#             return 2.0

#     cond_count2.code.print()
#     result = address_results(cond_count2)
#     assert isinstance(result, NotQubit)


def test_multi_return():
    @qasm2.main.add(py.tuple)
    def multi_return_cnt():
        ra = qasm2.qreg(3)
        rb = qasm2.qreg(4)
        return ra, 3.0, rb

    multi_return_cnt.code.print()
    fold(multi_return_cnt)
    _, result = address.run_analysis(multi_return_cnt)
    print(result)
    assert isinstance(result, AddressTuple)
    assert isinstance(result.data[0], AddressReg)
    assert isinstance(result.data[1], NotQubit)
    assert isinstance(result.data[2], AddressReg)


def test_list():
    @qasm2.main.add(ilist)
    def list_count_analy():
        ra = qasm2.qreg(3)
        rb = qasm2.qreg(4)
        f = [ra[0], ra[1], rb[0]]
        return f

    list_count_analy.code.print()
    _, ret = address.run_analysis(list_count_analy)
    print(ret)
    assert isinstance(ret, AddressTuple)
    assert isinstance(ret.data[0], AddressQubit) and ret.data[0].data == 0
    assert isinstance(ret.data[1], AddressQubit) and ret.data[1].data == 1
    assert isinstance(ret.data[2], AddressQubit) and ret.data[2].data == 3


def test_tuple_qubits():
    @qasm2.main.add(py.tuple)
    def list_count_analy2():
        ra = qasm2.qreg(3)
        rb = qasm2.qreg(4)

        f = (ra[0], ra[1], rb[0])
        return f

    list_count_analy2.code.print()
    _, ret = address.run_analysis(list_count_analy2)
    assert isinstance(ret, AddressTuple)
    assert isinstance(ret.data[0], AddressQubit) and ret.data[0].data == 0
    assert isinstance(ret.data[1], AddressQubit) and ret.data[1].data == 1
    assert isinstance(ret.data[2], AddressQubit) and ret.data[2].data == 3


# NOTE: invalid QASM2 program, use this test for future
# def test_tuple_qubits_tuple_add():
#     @qasm2.main.add(py.tuple)
#     def list_count_analy3():
#         ra = qasm2.qreg(3)
#         rb = qasm2.qreg(4)

#         f = (ra[0], ra[1], rb[0])
#         g = (ra[1], ra[0], rb[3])
#         return f + g

#     result = address_results(list_count_analy3)
#     list_count_analy3.print(analysis=address.results)
#     assert isinstance(result, AddressTuple)
#     assert len(result.data) == 6
#     assert isinstance(result.data[0], AddressQubit) and result.data[0].data == 0
#     assert isinstance(result.data[1], AddressQubit) and result.data[1].data == 1
#     assert isinstance(result.data[2], AddressQubit) and result.data[2].data == 3
#     assert isinstance(result.data[3], AddressQubit) and result.data[3].data == 1
#     assert isinstance(result.data[4], AddressQubit) and result.data[4].data == 0
#     assert isinstance(result.data[5], AddressQubit) and result.data[5].data == 6


def test_alias():

    @qasm2.main
    def test_alias():
        ra = qasm2.qreg(3)

        f = ra[0]
        g = f
        h = g

        return h

    test_alias.code.print()
    fold(test_alias)
    _, ret = address.run_analysis(test_alias)
    assert isinstance(ret, AddressQubit)
    assert ret.data == 0



from kirin.decl import statement

from .base import SingleQubitGate
from .._dialect import dialect


# Pauli Gates
# -----------------------------------
@statement(dialect=dialect)
class Identity(SingleQubitGate):
    name = "I"


@statement(dialect=dialect)
class X(SingleQubitGate):
    name = "X"


@statement(dialect=dialect)
class Y(SingleQubitGate):
    name = "Y"


@statement(dialect=dialect)
class Z(SingleQubitGate):
    name = "Z"


# Single Qubit Clifford Gates
# ---------------------------------------
@statement(dialect=dialect)
class H(SingleQubitGate):
    name = "H"


@statement(dialect=dialect)
class S(SingleQubitGate):
    name = "S"


@statement(dialect=dialect)
class SqrtX(SingleQubitGate):
    name = "SQRT_X"


@statement(dialect=dialect)
class SqrtY(SingleQubitGate):
    name = "SQRT_Y"


@statement(dialect=dialect)
class SqrtZ(SingleQubitGate):
    name = "SQRT_Z"



from typing import List
from dataclasses import dataclass

from kirin import ir
from bloqade import qasm2
from kirin.rewrite import abc, result
from kirin.dialects import py
from bloqade.analysis import address
from bloqade.qasm2.dialects import glob


@dataclass
class GlobalToUOpRule(abc.RewriteRule):
    address_regs: List[address.AddressReg]
    address_reg_ssas: List[ir.SSAValue]

    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        if node.dialect == glob.dialect:
            return getattr(self, f"rewrite_{node.name}")(node)

        return result.RewriteResult()

    def rewrite_ugate(self, node: ir.Statement):
        assert isinstance(node, glob.UGate)

        # if there's no register even found, just give up
        if not self.address_regs:
            return result.RewriteResult()

        for address_reg, address_reg_ssa in zip(
            self.address_regs, self.address_reg_ssas
        ):

            for qubit_idx in address_reg.data:

                qubit_idx = py.constant.Constant(value=qubit_idx)

                qubit_stmt = qasm2.core.QRegGet(
                    reg=address_reg_ssa, idx=qubit_idx.result
                )
                qubit_ssa = qubit_stmt.result

                ugate_node = qasm2.uop.UGate(
                    qarg=qubit_ssa, theta=node.theta, phi=node.phi, lam=node.lam
                )

                qubit_idx.insert_before(node)
                qubit_stmt.insert_before(node)
                ugate_node.insert_after(node)

        node.delete()

        return result.RewriteResult(has_done_something=True)



"""Rewrite py dialects into qasm dialects."""

from kirin import ir
from kirin.passes import Pass
from kirin.rewrite import Walk, Fixpoint
from kirin.dialects import py, math
from kirin.rewrite.abc import RewriteRule
from kirin.rewrite.result import RewriteResult
from bloqade.qasm2.dialects import core, expr


class _Py2QASM(RewriteRule):
    """Rewrite py dialects into qasm dialects."""

    UNARY_OPS = {
        py.USub: expr.Neg,
        math.sin: expr.Sin,
        math.cos: expr.Cos,
        math.tan: expr.Tan,
        math.exp: expr.Exp,
        math.sqrt: expr.Sqrt,
    }

    BINARY_OPS = {
        py.Add: expr.Add,
        py.Sub: expr.Sub,
        py.Mult: expr.Mul,
        py.Div: expr.Div,
        py.Pow: expr.Pow,
    }

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if isinstance(node, py.Constant):
            if isinstance(node.value, int):
                node.replace_by(expr.ConstInt(value=node.value))
                return RewriteResult(has_done_something=True)
            elif isinstance(node.value, float):
                node.replace_by(expr.ConstFloat(value=node.value))
                return RewriteResult(has_done_something=True)
        elif isinstance(node, py.BinOp):
            if (pystmt := self.BINARY_OPS.get(type(node))) is not None:
                node.replace_by(pystmt(node.lhs, node.rhs))
                return RewriteResult(has_done_something=True)
        elif isinstance(node, py.UnaryOp):
            if (pystmt := self.UNARY_OPS.get(type(node))) is not None:
                node.replace_by(pystmt(node.value))
                return RewriteResult(has_done_something=True)
        elif isinstance(node, py.cmp.Eq):
            node.replace_by(core.CRegEq(node.lhs, node.rhs))
            return RewriteResult(has_done_something=True)
        elif isinstance(node, py.assign.Alias):
            node.result.replace_by(node.value)
            node.delete()
            return RewriteResult(has_done_something=True)
        return RewriteResult()


class Py2QASM(Pass):

    def unsafe_run(self, mt: ir.Method) -> RewriteResult:
        return Fixpoint(Walk(_Py2QASM())).rewrite(mt.code)



import pathlib

from rich.console import Console

from . import ast as ast
from .build import Build
from .print import Printer as Printer
from .parser import qasm2_parser as lark_parser
from .visitor import Visitor as Visitor


def loads(txt: str):
    raw = lark_parser.parse(txt)
    return Build().build_mainprogram(raw)


def loadfile(file: str | pathlib.Path):
    with open(file) as f:
        return loads(f.read())


def pprint(node: ast.Node, *, console: Console | None = None):
    if console:
        return Printer(console).visit(node)
    else:
        Printer().visit(node)


def spprint(node: ast.Node, *, console: Console | None = None):
    if console:
        printer = Printer(console)
    else:
        printer = Printer()

    with printer.string_io() as stream:
        printer.visit(node)
        return stream.getvalue()



from kirin import ir, types
from kirin.decl import info, statement
from bloqade.qasm2.types import BitType, CRegType, QRegType, QubitType

from ._dialect import dialect


@statement(dialect=dialect)
class QRegNew(ir.Statement):
    """Create a new quantum register."""

    name = "qreg.new"
    traits = frozenset({ir.FromPythonCall()})
    n_qubits: ir.SSAValue = info.argument(types.Int)
    """n_qubits: The number of qubits in the register."""
    result: ir.ResultValue = info.result(QRegType)
    """A new quantum register with n_qubits set to |0>."""


@statement(dialect=dialect)
class CRegNew(ir.Statement):
    """Create a new classical register."""

    name = "creg.new"
    traits = frozenset({ir.FromPythonCall()})
    n_bits: ir.SSAValue = info.argument(types.Int)
    """n_bits (Int): The number of bits in the register."""
    result: ir.ResultValue = info.result(CRegType)
    """result (CReg): The new classical register with all bits set to 0."""


@statement(dialect=dialect)
class Reset(ir.Statement):
    """Reset a qubit to the |0> state."""

    name = "reset"
    traits = frozenset({ir.FromPythonCall()})
    qarg: ir.SSAValue = info.argument(QubitType)
    """qarg (Qubit): The qubit to reset."""


@statement(dialect=dialect)
class Measure(ir.Statement):
    """Measure a qubit and store the result in a bit."""

    name = "measure"
    traits = frozenset({ir.FromPythonCall()})
    qarg: ir.SSAValue = info.argument(QubitType)
    """qarg (Qubit): The qubit to measure."""
    carg: ir.SSAValue = info.argument(BitType)
    """carg (Bit): The bit to store the result in."""


@statement(dialect=dialect)
class CRegEq(ir.Statement):
    """Check if two classical registers are equal."""

    name = "eq"
    traits = frozenset({ir.Pure(), ir.FromPythonCall()})
    lhs: ir.SSAValue = info.argument(types.Int | CRegType | BitType)
    """lhs (CReg): The first register."""
    rhs: ir.SSAValue = info.argument(types.Int | CRegType | BitType)
    """rhs (CReg): The second register."""
    result: ir.ResultValue = info.result(types.Bool)
    """result (bool): True if the registers are equal, False otherwise."""


@statement(dialect=dialect)
class QRegGet(ir.Statement):
    """Get a qubit from a quantum register."""

    name = "qreg.get"
    traits = frozenset({ir.FromPythonCall(), ir.Pure()})
    reg: ir.SSAValue = info.argument(QRegType)
    """reg (QReg): The quantum register."""
    idx: ir.SSAValue = info.argument(types.Int)
    """idx (Int): The index of the qubit in the register."""
    result: ir.ResultValue = info.result(QubitType)
    """result (Qubit): The qubit at position `idx`."""


@statement(dialect=dialect)
class CRegGet(ir.Statement):
    """Get a bit from a classical register."""

    name = "creg.get"
    traits = frozenset({ir.FromPythonCall(), ir.Pure()})
    reg: ir.SSAValue = info.argument(CRegType)
    """reg (CReg): The classical register."""
    idx: ir.SSAValue = info.argument(types.Int)
    """idx (Int): The index of the bit in the register."""
    result: ir.ResultValue = info.result(BitType)
    """result (Bit): The bit at position `idx`."""



from dataclasses import field, dataclass

from kirin.passes import Pass
from kirin.rewrite import (
    Walk,
    Chain,
    Inline,
    Fixpoint,
    WrapConst,
    Call2Invoke,
    ConstantFold,
    CFGCompactify,
    InlineGetItem,
    InlineGetField,
    DeadCodeElimination,
    CommonSubexpressionElimination,
)
from kirin.analysis import const
from kirin.dialects import scf, ilist
from kirin.ir.method import Method
from kirin.rewrite.abc import RewriteResult
from bloqade.qasm2.dialects import expr


@dataclass
class QASM2Fold(Pass):
    """Fold pass for qasm2.extended"""

    constprop: const.Propagate = field(init=False)
    inline_gate_subroutine: bool = True

    def __post_init__(self):
        self.constprop = const.Propagate(self.dialects)

    def unsafe_run(self, mt: Method) -> RewriteResult:
        result = RewriteResult()
        frame, _ = self.constprop.run_analysis(mt)
        result = Walk(WrapConst(frame)).rewrite(mt.code).join(result)
        rule = Chain(
            ConstantFold(),
            Call2Invoke(),
            InlineGetField(),
            InlineGetItem(),
            DeadCodeElimination(),
            CommonSubexpressionElimination(),
        )
        result = Fixpoint(Walk(rule)).rewrite(mt.code).join(result)

        result = (
            Walk(
                Chain(
                    scf.unroll.PickIfElse(),
                    scf.unroll.ForLoop(),
                    scf.trim.UnusedYield(),
                )
            )
            .rewrite(mt.code)
            .join(result)
        )
        result = Walk(ilist.rewrite.Unroll()).rewrite(mt.code).join(result)

        def skip_scf(node):
            return isinstance(node, (scf.For, scf.IfElse))

        result = (
            Walk(
                Inline(
                    lambda x: (
                        True
                        if self.inline_gate_subroutine
                        else not isinstance(x, expr.GateFunction)
                    )
                ),
                skip=skip_scf,
            )
            .rewrite(mt.code)
            .join(result)
        )
        result = Fixpoint(CFGCompactify()).rewrite(mt.code).join(result)
        return result



"""QASM2 extension for global gates."""

from typing import Any

from kirin.dialects import ilist
from kirin.lowering import wraps

from .types import QReg
from .dialects import glob


@wraps(glob.UGate)
def u(
    theta: float, phi: float, lam: float, registers: ilist.IList[QReg, Any] | list
) -> None:
    """Apply a U gate to all qubits in the input registers.

    Args:
        theta (float): The angle theta.
        phi (float): The angle phi.
        lam (float): The angle lam.
        registers (IList[QReg] | list[QReg]): The registers to apply the gate to.

    """



from typing import TYPE_CHECKING, Union, Optional

from kirin import ir

if TYPE_CHECKING:
    from qbraid import QbraidProvider
    from qbraid.runtime import QbraidJob

from bloqade.qasm2.emit import QASM2


class qBraid:
    """qBraid target for Bloqade kernels.

    qBraid target that accepts a Bloqade kernel and submits the kernel to the QuEra simulator hosted on qBraid. A `QbraidJob` is obtainable
    that then lets you query the status of the submitted program on the simulator as well
    as obtain results.

    """

    def __init__(
        self,
        *,
        main_target: ir.DialectGroup | None = None,
        gate_target: ir.DialectGroup | None = None,
        provider: "QbraidProvider",  # inject externally for easier mocking
        qelib1: bool = True,
    ) -> None:
        """Initialize the qBraid target.

        Args:
            main_target (ir.DialectGroup | None):
                The dialects that were used in the definition of the kernel. This is used to
                generate the correct header for the resulting QASM2 AST. Argument defaults to `None`.
                Internally set to the `qasm2.main` group of dialects.
            gate_target (ir.DialectGroup | None):
                The dialects involved in defining any custom gates in the kernel. Argument defaults to `None`.
                Internally set to the `qasm2.gate` group of dialects.
            provider (QbraidProvider):
                Qbraid-provided object to allow submission of the kernel to the QuEra simulator.
            qelib1 (bool):
                Include the `include "qelib1.inc"` line in the resulting QASM2 AST that's
                submitted to qBraid. Defaults to `True`.
        """

        from bloqade import qasm2

        self.main_target = main_target or qasm2.main
        self.gate_target = gate_target or qasm2.gate
        self.qelib1 = qelib1
        self.provider = provider

    def emit(
        self,
        method: ir.Method,
        shots: Optional[int] = None,
        tags: Optional[dict[str, str]] = None,
    ) -> Union["QbraidJob", list["QbraidJob"]]:
        """Submit the Bloqade kernel to the QuEra simulator on qBraid.

        Args:
            method (ir.Method):
                The kernel to submit to qBraid.
            shots: (Optional[int]):
                Number of times to run the kernel. Defaults to None.
            tags: (Optional[dict[str,str]]):
                A dictionary of tags to associate with the Job.

        Returns:
            Union[QbraidJob, list[QbraidJob]]:
                An object you can query for the status of your submission as well as
                obtain simulator results from.
        """

        # Convert method to QASM2 string
        qasm2_emitter = QASM2(
            main_target=self.main_target,
            gate_target=self.gate_target,
            qelib1=self.qelib1,
        )
        qasm2_prog = qasm2_emitter.emit_str(method)

        # Submit the QASM2 string to the qBraid simulator
        quera_qasm_simulator = self.provider.get_device("quera_qasm_simulator")

        return quera_qasm_simulator.run(qasm2_prog, shots=shots, tags=tags)



from . import lowering as lowering
from .emit import EmitStimAuxMethods as EmitStimAuxMethods
from .stmts import *  # noqa F403
from .types import (
    RecordType as RecordType,
    PauliString as PauliString,
    RecordResult as RecordResult,
    PauliStringType as PauliStringType,
)
from .interp import StimAuxMethods as StimAuxMethods
from ._dialect import dialect as dialect



from bloqade import stim

from .base import codegen


def test_x():

    @stim.main
    def test_x():
        stim.x(targets=(0, 1, 2, 3), dagger=False)

    test_x.print()
    out = codegen(test_x)

    assert out.strip() == "X 0 1 2 3"

    @stim.main
    def test_x_dag():
        stim.x(targets=(0, 1, 2, 3), dagger=True)

    out = codegen(test_x_dag)

    assert out.strip() == "X 0 1 2 3"


def test_y():

    @stim.main
    def test_y():
        stim.y(targets=(0, 1, 2, 3), dagger=False)

    test_y.print()
    out = codegen(test_y)

    assert out.strip() == "Y 0 1 2 3"

    @stim.main
    def test_y_dag():
        stim.y(targets=(0, 1, 2, 3), dagger=True)

    out = codegen(test_y_dag)

    assert out.strip() == "Y 0 1 2 3"



from . import impls as impls
from .lattice import (
    Address as Address,
    NotQubit as NotQubit,
    AddressReg as AddressReg,
    AnyAddress as AnyAddress,
    AddressQubit as AddressQubit,
    AddressTuple as AddressTuple,
)
from .analysis import AddressAnalysis as AddressAnalysis



from kirin import ir, types
from kirin.decl import info, statement

from ..types import RecordType, PauliStringType
from .._dialect import dialect

PyNum = types.Union(types.Int, types.Float)


@statement(dialect=dialect)
class GetRecord(ir.Statement):
    name = "get_rec"
    traits = frozenset({ir.FromPythonCall()})
    id: ir.SSAValue = info.argument(type=types.Int)
    result: ir.ResultValue = info.result(type=RecordType)


@statement(dialect=dialect)
class Detector(ir.Statement):
    name = "detector"
    traits = frozenset({ir.FromPythonCall()})
    coord: tuple[ir.SSAValue, ...] = info.argument(PyNum)
    targets: tuple[ir.SSAValue, ...] = info.argument(RecordType)


@statement(dialect=dialect)
class ObservableInclude(ir.Statement):
    name = "obs.include"
    traits = frozenset({ir.FromPythonCall()})
    idx: ir.SSAValue = info.argument(type=types.Int)
    targets: tuple[ir.SSAValue, ...] = info.argument(RecordType)


@statement(dialect=dialect)
class Tick(ir.Statement):
    name = "tick"
    traits = frozenset({ir.FromPythonCall()})


@statement(dialect=dialect)
class NewPauliString(ir.Statement):
    name = "new_pauli_string"
    traits = frozenset({ir.FromPythonCall()})
    string: tuple[ir.SSAValue, ...] = info.argument(types.String)
    flipped: tuple[ir.SSAValue, ...] = info.argument(types.Bool)
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)
    result: ir.ResultValue = info.result(type=PauliStringType)
